# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the BSD license.
# See LICENSE in the root of the repository for full licensing details.
"""
Unit tests for :func:`iris_grib._save_rules.set_fixed_surfaces`.

"""

# Import iris_grib.tests first so that some things can be initialised before
# importing anything else.
import iris_grib.tests as tests

from unittest import mock

import eccodes
from eccodes import CODES_MISSING_LONG as GRIB_MISSING_LONG
import numpy as np

import iris.cube
import iris.coords
from iris.exceptions import TranslationError

from iris_grib._save_rules import set_fixed_surfaces


class Test(tests.IrisGribTest):
    def test_bounded_altitude_feet(self):
        cube = iris.cube.Cube([0])
        cube.add_aux_coord(
            iris.coords.AuxCoord(
                1500.0,
                long_name="altitude",
                units="ft",
                bounds=np.array([1000.0, 2000.0]),
            )
        )
        grib = eccodes.codes_grib_new_from_samples("GRIB2")
        set_fixed_surfaces(cube, grib)
        self.assertEqual(
            eccodes.codes_get_double(grib, "scaledValueOfFirstFixedSurface"), 305.0
        )  # precise ~304.8
        self.assertEqual(
            eccodes.codes_get_double(grib, "scaledValueOfSecondFixedSurface"), 610.0
        )  # precise ~609.6
        self.assertEqual(eccodes.codes_get_long(grib, "typeOfFirstFixedSurface"), 102)
        self.assertEqual(eccodes.codes_get_long(grib, "typeOfSecondFixedSurface"), 102)

    def test_theta_level(self):
        cube = iris.cube.Cube([0])
        cube.add_aux_coord(
            iris.coords.AuxCoord(
                230.0,
                standard_name="air_potential_temperature",
                units="K",
                attributes={"positive": "up"},
                bounds=np.array([220.0, 240.0]),
            )
        )
        grib = eccodes.codes_grib_new_from_samples("GRIB2")
        set_fixed_surfaces(cube, grib)
        self.assertEqual(
            eccodes.codes_get_double(grib, "scaledValueOfFirstFixedSurface"), 220.0
        )
        self.assertEqual(
            eccodes.codes_get_double(grib, "scaledValueOfSecondFixedSurface"), 240.0
        )
        self.assertEqual(eccodes.codes_get_long(grib, "typeOfFirstFixedSurface"), 107)
        self.assertEqual(eccodes.codes_get_long(grib, "typeOfSecondFixedSurface"), 107)

    def test_depth(self):
        cube = iris.cube.Cube([0])
        cube.add_aux_coord(
            iris.coords.AuxCoord(
                1,
                long_name="depth",
                units="m",
                bounds=np.array([0.0, 2]),
                attributes={"positive": "down"},
            )
        )
        grib = eccodes.codes_grib_new_from_samples("GRIB2")
        set_fixed_surfaces(cube, grib)
        self.assertEqual(
            eccodes.codes_get_double(grib, "scaledValueOfFirstFixedSurface"), 0.0
        )
        self.assertEqual(
            eccodes.codes_get_double(grib, "scaledValueOfSecondFixedSurface"), 2
        )
        self.assertEqual(eccodes.codes_get_long(grib, "typeOfFirstFixedSurface"), 106)
        self.assertEqual(eccodes.codes_get_long(grib, "typeOfSecondFixedSurface"), 106)

    @mock.patch.object(eccodes, "codes_set")
    def test_altitude_point(self, mock_set):
        grib = None
        cube = iris.cube.Cube([1, 2, 3, 4, 5])
        cube.add_aux_coord(iris.coords.AuxCoord([12345], "altitude", units="m"))

        set_fixed_surfaces(cube, grib)

        mock_set.assert_any_call(grib, "typeOfFirstFixedSurface", 102)
        mock_set.assert_any_call(grib, "scaleFactorOfFirstFixedSurface", 0)
        mock_set.assert_any_call(grib, "scaledValueOfFirstFixedSurface", 12345)
        mock_set.assert_any_call(grib, "typeOfSecondFixedSurface", 255)
        mock_set.assert_any_call(
            grib, "scaleFactorOfSecondFixedSurface", GRIB_MISSING_LONG
        )
        mock_set.assert_any_call(
            grib, "scaledValueOfSecondFixedSurface", GRIB_MISSING_LONG
        )

    @mock.patch.object(eccodes, "codes_set")
    def test_height_point(self, mock_set):
        grib = None
        cube = iris.cube.Cube([1, 2, 3, 4, 5])
        cube.add_aux_coord(iris.coords.AuxCoord([12345], "height", units="m"))

        set_fixed_surfaces(cube, grib)

        mock_set.assert_any_call(grib, "typeOfFirstFixedSurface", 103)
        mock_set.assert_any_call(grib, "scaleFactorOfFirstFixedSurface", 0)
        mock_set.assert_any_call(grib, "scaledValueOfFirstFixedSurface", 12345)
        mock_set.assert_any_call(grib, "typeOfSecondFixedSurface", 255)
        mock_set.assert_any_call(
            grib, "scaleFactorOfSecondFixedSurface", GRIB_MISSING_LONG
        )
        mock_set.assert_any_call(
            grib, "scaledValueOfSecondFixedSurface", GRIB_MISSING_LONG
        )

    def test_unknown_vertical_unbounded(self):
        cube = iris.cube.Cube([0])
        cube.add_aux_coord(
            iris.coords.AuxCoord([1], attributes={"GRIB_fixed_surface_type": 151})
        )
        grib = eccodes.codes_grib_new_from_samples("GRIB2")
        set_fixed_surfaces(cube, grib)

        self.assertEqual(eccodes.codes_get_long(grib, "typeOfFirstFixedSurface"), 151)
        self.assertEqual(
            eccodes.codes_get_double(grib, "scaledValueOfFirstFixedSurface"), 1
        )
        self.assertEqual(
            eccodes.codes_get_double(grib, "scaleFactorOfFirstFixedSurface"), 0
        )
        self.assertEqual(eccodes.codes_get_long(grib, "typeOfSecondFixedSurface"), 255)
        self.assertEqual(
            eccodes.codes_get_long(grib, "scaledValueOfSecondFixedSurface"),
            GRIB_MISSING_LONG,
        )
        self.assertEqual(
            eccodes.codes_get_long(grib, "scaleFactorOfSecondFixedSurface"),
            GRIB_MISSING_LONG,
        )

    def test_unknown_vertical_bounded(self):
        cube = iris.cube.Cube([0])
        cube.add_aux_coord(
            iris.coords.AuxCoord(
                [700],
                bounds=np.array([900.0, 500.0]),
                attributes={"GRIB_fixed_surface_type": 108},
            )
        )
        grib = eccodes.codes_grib_new_from_samples("GRIB2")
        set_fixed_surfaces(cube, grib)

        self.assertEqual(eccodes.codes_get_long(grib, "typeOfFirstFixedSurface"), 108)
        self.assertEqual(
            eccodes.codes_get_double(grib, "scaledValueOfFirstFixedSurface"), 900
        )
        self.assertEqual(
            eccodes.codes_get_double(grib, "scaleFactorOfFirstFixedSurface"), 0
        )
        self.assertEqual(eccodes.codes_get_long(grib, "typeOfSecondFixedSurface"), 108)
        self.assertEqual(
            eccodes.codes_get_long(grib, "scaledValueOfSecondFixedSurface"), 500
        )
        self.assertEqual(
            eccodes.codes_get_long(grib, "scaleFactorOfSecondFixedSurface"), 0
        )

    def test_multiple_unknown_vertical_coords(self):
        grib = None
        cube = iris.cube.Cube([0])
        cube.add_aux_coord(
            iris.coords.AuxCoord([1], attributes={"GRIB_fixed_surface_type": 151})
        )
        cube.add_aux_coord(
            iris.coords.AuxCoord(
                [450],
                bounds=np.array([900.0, 0.0]),
                attributes={"GRIB_fixed_surface_type": 108},
            )
        )
        msg = r"coordinates were found of fixed surface type: \[151, 108\]"
        with self.assertRaisesRegex(ValueError, msg):
            set_fixed_surfaces(cube, grib)

    def test_unhandled_vertical_axis(self):
        grib = None
        cube = iris.cube.Cube([0])
        cube.add_aux_coord(iris.coords.AuxCoord([450], attributes={"positive": "up"}))
        msg = (
            r"vertical-axis coordinate\(s\) \('unknown'\) are not "
            "recognised or handled."
        )
        with self.assertRaisesRegex(TranslationError, msg):
            set_fixed_surfaces(cube, grib)

    @mock.patch.object(eccodes, "codes_set")
    def test_no_vertical(self, mock_set):
        grib = None
        cube = iris.cube.Cube([1, 2, 3, 4, 5])
        set_fixed_surfaces(cube, grib)
        mock_set.assert_any_call(grib, "typeOfFirstFixedSurface", 1)
        mock_set.assert_any_call(grib, "scaleFactorOfFirstFixedSurface", 0)
        mock_set.assert_any_call(grib, "scaledValueOfFirstFixedSurface", 0)
        mock_set.assert_any_call(grib, "typeOfSecondFixedSurface", 255)
        mock_set.assert_any_call(
            grib, "scaleFactorOfSecondFixedSurface", GRIB_MISSING_LONG
        )
        mock_set.assert_any_call(
            grib, "scaledValueOfSecondFixedSurface", GRIB_MISSING_LONG
        )


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the BSD license.
# See LICENSE in the root of the repository for full licensing details.
"""
Unit tests for :func:`iris_grib._save_rules.product_definition_template_6`

"""

# Import iris_grib.tests first so that some things can be initialised before
# importing anything else.
import iris_grib.tests as tests

from unittest import mock

from cf_units import Unit
import eccodes

from iris.coords import DimCoord
import iris.tests.stock as stock

from iris_grib._save_rules import product_definition_template_6


class TestRealizationIdentifier(tests.IrisGribTest):
    def setUp(self):
        self.cube = stock.lat_lon_cube()
        # Rename cube to avoid warning about unknown discipline/parameter.
        self.cube.rename("air_temperature")
        coord = DimCoord(
            [45], "time", units=Unit("days since epoch", calendar="standard")
        )
        self.cube.add_aux_coord(coord)

    @mock.patch.object(eccodes, "codes_set")
    def test_percentile(self, mock_set):
        cube = self.cube
        coord = DimCoord(10, long_name="percentile", units="%")
        cube.add_aux_coord(coord)

        product_definition_template_6(cube, mock.sentinel.grib)
        mock_set.assert_any_call(
            mock.sentinel.grib, "productDefinitionTemplateNumber", 6
        )
        mock_set.assert_any_call(mock.sentinel.grib, "percentileValue", 10)

    @mock.patch.object(eccodes, "codes_set")
    def test_multiple_percentile_values(self, mock_set):
        cube = self.cube
        coord = DimCoord([8, 9, 10], long_name="percentile", units="%")
        cube.add_aux_coord(coord, 0)

        msg = "'percentile' coordinate with one point is required"
        with self.assertRaisesRegex(ValueError, msg):
            product_definition_template_6(cube, mock.sentinel.grib)


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the BSD license.
# See LICENSE in the root of the repository for full licensing details.
"""
Perform test automation with nox.

For further details, see https://nox.thea.codes/en/stable/#

"""

from contextlib import contextmanager
import hashlib
import os
from pathlib import Path

import nox
from nox.logger import logger

#: Default to reusing any pre-existing nox environments.
nox.options.reuse_existing_virtualenvs = True

#: Name of the package to test.
PACKAGE = str("iris_grib")

#: Cirrus-CI environment variable hook.
PY_VER = os.environ.get("PY_VER", ["3.10", "3.11", "3.12"])
IRIS_SOURCE = os.environ.get("IRIS_SOURCE", ["source", "conda-forge"])

#: Default cartopy cache directory.
CARTOPY_CACHE_DIR = os.environ.get("HOME") / Path(".local/share/cartopy")


def _cache_cartopy(session: nox.sessions.Session) -> None:
    """
    Determine whether to cache the cartopy natural earth shapefiles.

    Parameters
    ----------
    session: object
        A `nox.sessions.Session` object.

    """
    if not CARTOPY_CACHE_DIR.is_dir():
        session.run(
            "python",
            "-c",
            "import cartopy; cartopy.io.shapereader.natural_earth()",
        )


def _write_iris_config(session: nox.sessions.Session) -> None:
    """
    Add test data dir and libudunits2.so to iris config.

    test data dir is set from session pos args. i.e. can be
    configured by passing in on the command line:
        nox --session tests -- --test-data-dir $TEST_DATA_DIR/test_data

    Parameters
    ----------
    session: object
        A `nox.sessions.Session` object.

    """
    try:
        test_data_dir = session.posargs[session.posargs.index("--test-data-dir") + 1]
    except Exception:
        test_data_dir = ""

    iris_config_file = os.path.join(
        session.virtualenv.location,
        "lib",
        f"python{session.python}",
        "site-packages",
        "iris",
        "etc",
        "site.cfg",
    )
    iris_config = f"""
[Resources]
test_data_dir = {test_data_dir}
[System]
udunits2_path = {
    os.path.join(session.virtualenv.location,'lib', 'libudunits2.so')
}
"""

    print("Iris config\n-----------")
    print(iris_config)

    with open(iris_config_file, "w") as f:
        f.write(iris_config)


def _session_lockfile(session: nox.sessions.Session) -> Path:
    """
    Return the path of the session lockfile for the relevant python string
    e.g ``py38``.

    """
    lockfile_name = f"py{session.python.replace('.', '')}-linux-64.lock"
    return Path("requirements/locks") / lockfile_name


def _file_content(file_path: Path) -> str:
    with file_path.open("r") as file:
        return file.read()


def _session_cachefile(session: nox.sessions.Session) -> Path:
    """Return the path of the session lockfile cache."""
    tmp_dir = Path(session.create_tmp())
    cache = tmp_dir / _session_lockfile(session).name
    return cache


def _venv_populated(session: nox.sessions.Session) -> bool:
    """
    Return True if the Conda venv has been created and the list of packages in
    the lockfile installed.

    """
    return _session_cachefile(session).is_file()


def _venv_changed(session: nox.sessions.Session) -> bool:
    """
    Return True if the installed session is different to that specified in the
    lockfile.

    """
    result = False
    if _venv_populated(session):
        expected = _file_content(_session_lockfile(session))
        actual = _file_content(_session_cachefile(session))
        result = actual != expected
    return result


def _install_and_cache_venv(session: nox.sessions.Session) -> None:
    """
    Install and cache the nox session environment.
    This consists of saving a hexdigest (sha256) of the associated
    Conda lock file.

    Parameters
    ----------
    session: object
        A `nox.sessions.Session` object.

    """
    lockfile = _session_lockfile(session)
    session.conda_install(f"--file={lockfile}")

    with open(lockfile, "rb") as fi:
        hexdigest = hashlib.sha256(fi.read()).hexdigest()
    with _session_cachefile(session).open("w") as cachefile:
        cachefile.write(hexdigest)


@contextmanager
def prepare_venv(
    session: nox.sessions.Session, iris_source: str = "conda-forge"
) -> None:
    """
    Create and cache the nox session conda environment, and additionally
    provide conda environment package details and info.

    Note that, iris-grib is installed into the environment using pip.

    Parameters
    ----------
    session: object
        A `nox.sessions.Session` object.

    iris_source: str
        Determines where Iris was sourced from. Either 'conda-forge' (the
        default), or 'source' which refers to the Iris main branch.

    Notes
    -----
    See
      - https://github.com/theacodes/nox/issues/346
      - https://github.com/theacodes/nox/issues/260

    """
    venv_dir = session.virtualenv.location_name

    if not _venv_populated(session):
        # Environment has been created but packages not yet installed.
        # Populate the environment from the lockfile.
        logger.debug(f"Populating conda env: {venv_dir}")
        _install_and_cache_venv(session)

    elif _venv_changed(session):
        # Destroy the environment and rebuild it.
        logger.debug(f"Lockfile changed. Recreating conda env: {venv_dir}")
        _reuse_original = session.virtualenv.reuse_existing
        session.virtualenv.reuse_existing = False
        session.virtualenv.create()
        _install_and_cache_venv(session)
        session.virtualenv.reuse_existing = _reuse_original

    logger.debug(f"Environment up to date: {venv_dir}")

    if iris_source == "source":
        # get latest iris
        iris_dir = f"{session.create_tmp()}/iris"

        if os.path.exists(iris_dir):
            # cached.  update by pulling from origin/master
            session.run(
                "git",
                "-C",
                iris_dir,
                "pull",
                "origin",
                "main",
                external=True,  # use git from host environment
            )
        else:
            session.run(
                "git",
                "clone",
                "https://github.com/scitools/iris.git",
                iris_dir,
                external=True,
            )
        session.install(iris_dir, "--no-deps")

    _cache_cartopy(session)
    _write_iris_config(session)

    # Install the iris-grib source in develop mode.
    session.install("--no-deps", "--editable", ".")

    # Determine whether verbose diagnostics have been requested
    # from the command line.
    verbose = "-v" in session.posargs or "--verbose" in session.posargs

    if verbose:
        session.run("conda", "info")
        session.run("conda", "list", f"--prefix={venv_dir}")
        session.run(
            "conda",
            "list",
            f"--prefix={venv_dir}",
            "--explicit",
        )


@nox.session(python=PY_VER, venv_backend="conda")
@nox.parametrize("iris_source", IRIS_SOURCE)
def tests(session: nox.sessions.Session, iris_source: str):
    """
    Perform iris-grib tests against release and development versions of iris.

    Parameters
    ----------
    session: object
        A `nox.sessions.Session` object.

    iris_source: str
        Either 'conda-forge' if using Iris from conda-forge, or 'source' if
        installing Iris from the Iris' main branch.

    """
    prepare_venv(session, iris_source)

    session.run("python", "-m", "eccodes", "selfcheck")

    run_args = [
        "pytest",
        "--pyargs",
        "iris_grib",
    ]

    if "-c" in session.posargs or "--coverage" in session.posargs:
        run_args.extend(["--cov=iris_grib", "--cov-report=xml"])
    session.run(*run_args)


@nox.session(python=PY_VER, venv_backend="conda")
def doctest(session: nox.sessions.Session):
    """
    Perform iris-grib doc-tests.

    Parameters
    ----------
    session: object
        A `nox.sessions.Session` object.

    """
    prepare_venv(session)
    session.cd("docs")
    session.run(
        "make",
        "clean",
        "html",
        external=True,
    )
    session.run(
        "make",
        "doctest",
        external=True,
    )


@nox.session(python=PY_VER, venv_backend="conda")
def linkcheck(session: nox.sessions.Session):
    """
    Perform iris-grib doc link check.

    Parameters
    ----------
    session: object
        A `nox.sessions.Session` object.

    """
    prepare_venv(session)
    session.cd("docs")
    session.run(
        "make",
        "clean",
        "html",
        external=True,
    )
    session.run(
        "make",
        "linkcheck",
        external=True,
    )



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the BSD license.
# See LICENSE in the root of the repository for full licensing details.
"""
Unit tests for :func:`iris_grib._save_rules.set_time_increment`

"""

# Import iris_grib.tests first so that some things can be initialised before
# importing anything else.
import iris_grib.tests as tests

from unittest import mock

import eccodes

from iris.coords import CellMethod

from iris_grib._save_rules import set_time_increment


class Test(tests.IrisGribTest):
    @mock.patch.object(eccodes, "codes_set")
    def test_no_intervals(self, mock_set):
        cell_method = CellMethod("sum", "time")
        set_time_increment(cell_method, mock.sentinel.grib)
        mock_set.assert_any_call(
            mock.sentinel.grib, "indicatorOfUnitForTimeIncrement", 255
        )
        mock_set.assert_any_call(mock.sentinel.grib, "timeIncrement", 0)

    @mock.patch.object(eccodes, "codes_set")
    def test_area(self, mock_set):
        cell_method = CellMethod("sum", "area", "25 km")
        set_time_increment(cell_method, mock.sentinel.grib)
        mock_set.assert_any_call(
            mock.sentinel.grib, "indicatorOfUnitForTimeIncrement", 255
        )
        mock_set.assert_any_call(mock.sentinel.grib, "timeIncrement", 0)

    @mock.patch.object(eccodes, "codes_set")
    def test_multiple_intervals(self, mock_set):
        cell_method = CellMethod("sum", "time", ("1 hour", "24 hour"))
        set_time_increment(cell_method, mock.sentinel.grib)
        mock_set.assert_any_call(
            mock.sentinel.grib, "indicatorOfUnitForTimeIncrement", 255
        )
        mock_set.assert_any_call(mock.sentinel.grib, "timeIncrement", 0)

    @mock.patch.object(eccodes, "codes_set")
    def test_hr(self, mock_set):
        cell_method = CellMethod("sum", "time", "23 hr")
        set_time_increment(cell_method, mock.sentinel.grib)
        mock_set.assert_any_call(
            mock.sentinel.grib, "indicatorOfUnitForTimeIncrement", 1
        )
        mock_set.assert_any_call(mock.sentinel.grib, "timeIncrement", 23)

    @mock.patch.object(eccodes, "codes_set")
    def test_hour(self, mock_set):
        cell_method = CellMethod("sum", "time", "24 hour")
        set_time_increment(cell_method, mock.sentinel.grib)
        mock_set.assert_any_call(
            mock.sentinel.grib, "indicatorOfUnitForTimeIncrement", 1
        )
        mock_set.assert_any_call(mock.sentinel.grib, "timeIncrement", 24)

    @mock.patch.object(eccodes, "codes_set")
    def test_hours(self, mock_set):
        cell_method = CellMethod("sum", "time", "25 hours")
        set_time_increment(cell_method, mock.sentinel.grib)
        mock_set.assert_any_call(
            mock.sentinel.grib, "indicatorOfUnitForTimeIncrement", 1
        )
        mock_set.assert_any_call(mock.sentinel.grib, "timeIncrement", 25)

    @mock.patch.object(eccodes, "codes_set")
    def test_fractional_hours(self, mock_set):
        cell_method = CellMethod("sum", "time", "25.9 hours")
        with mock.patch("warnings.warn") as warn:
            set_time_increment(cell_method, mock.sentinel.grib)
        warn.assert_called_once_with(
            "Truncating floating point timeIncrement " "25.9 to integer value 25"
        )
        mock_set.assert_any_call(
            mock.sentinel.grib, "indicatorOfUnitForTimeIncrement", 1
        )
        mock_set.assert_any_call(mock.sentinel.grib, "timeIncrement", 25)


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the BSD license.
# See LICENSE in the root of the repository for full licensing details.
"""
Provide object to represent grib phenomena

For use as cube attributes, freely convertible to+from strings
"""

from __future__ import annotations
from dataclasses import dataclass
import re


def _invalid_edition(edition):
    msg = f"Invalid grib edition, {edition!r}, for GRIBcode : " "can only be 1 or 2."
    raise ValueError(msg)


def _invalid_nargs(args):
    nargs = len(args)
    msg = (
        f"Cannot create GRIBCode from {nargs} arguments, "
        f"GRIBCode({args!r}) : expects either 1 or 4 arguments."
    )
    raise ValueError(msg)


# Regexp to extract four integers from a string:
# - for four times ...
# - match any non-digits (including none) and discard
# - then match any digits (including none), and return as a "group"
_RE_PARSE_FOURNUMS = re.compile(4 * r"[^\d]*(\d*)")


def _fournums_from_gribcode_string(grib_param_string):
    parsed_ok = True
    # get the numbers..
    match_groups = _RE_PARSE_FOURNUMS.match(grib_param_string).groups()
    # N.B. always produces 4 "strings of digits", but some can be empty
    try:
        nums = [int(grp) for grp in match_groups]
    except ValueError:
        parsed_ok = False

    if not parsed_ok:
        msg = (
            "Invalid argument for GRIBCode creation, "
            '"GRIBCode({!r})" : '
            "requires 4 numbers, separated by non-numerals."
        )
        raise ValueError(msg.format(grib_param_string))

    return nums


def GRIBCode(edition, *args, **kwargs):
    """
    Make an object representing a specific Grib phenomenon identity.

    The class of the result, and the list of its properties, depend on whether
    'edition' is 1 or 2.

    One of :

    * GRIBCode(edition=1, table_version, centre_number, number)
    * GRIBCode(edition=2, discipline, category, number)

    Either provides a string representation, and supports creation from:
    keywords, another similar object; a tuple of numbers; or any string with 4
    separate decimal numbers in it.
    """
    if edition is None:
        _invalid_nargs(())

    # Convert single argument to *args
    if not args and not kwargs:
        # Convert to a string and extract 4 integers.
        # NOTE: this also allows input from a GRIBCode, or a plain tuple.
        edition_string = str(edition)
        edition, arg2, arg3, arg4 = _fournums_from_gribcode_string(edition_string)
        args = [arg2, arg3, arg4]

    # Check edition + select the relevant keywords for the edition
    if edition not in (1, 2):
        _invalid_edition(edition)

    # Choose which actual type we will return.  This also determines the
    # argument (keyword) names.
    instance_cls = {1: GRIBCode1, 2: GRIBCode2}[edition]

    # Convert all of (edition, *args) into **kwargs
    if not args:
        # Ignore that edition= is a required arg -- make it a kwarg
        kwargs["edition"] = edition
    else:
        # Include edition, which just makes it simpler
        args = tuple([edition] + list(args))
        nargs = len(args)
        if nargs != 4:
            _invalid_nargs(args)

        for i_arg, (arg, name) in enumerate(
            zip(args, instance_cls.argnames, strict=False)
        ):
            if name in kwargs:
                msg = (
                    f"Keyword {name!r}={kwargs[name]!r} "
                    f"is not compatible with a {i_arg + 1}th argument."
                )
                raise ValueError(msg)
            else:
                kwargs[name] = arg

    result = instance_cls(**kwargs)
    return result


@dataclass
class GenericConcreteGRIBCode:
    """
    Common behaviour for GRIBCode1 and GRIBCode2

    GRIBCode1 and GRIBCode2 inherit this, making both dataclasses.
    They contain different data properties.
    """

    def __init__(self, **kwargs):
        # Note : only support creation with kargs.  In GRIBCode(), any args
        #  get translated into kwargs
        # Check against "_edition", defined by the specific subclass.
        assert kwargs["edition"] == self._edition
        for key, value in kwargs.items():
            setattr(self, key, value)

    def _broken_repr(self):
        result = (
            f"<{self.__class__.__name__} with invalid content: " f"{self.__dict__}>"
        )
        return result

    def __str__(self):
        edition = self.edition
        try:
            # NB fallback to "invalid" if edition not one of (1, 2)
            format = {
                1: "GRIB{:1d}:t{:03d}c{:03d}n{:03d}",
                2: "GRIB{:1d}:d{:03d}c{:03d}n{:03d}",
            }[edition]
            arg_values = [getattr(self, argname) for argname in self.argnames]
            # NB fallback to "invalid" if format fails
            result = format.format(*arg_values)
        except Exception:
            # Invalid content somewhere : fall back on default repr
            result = self._broken_repr()

        return result

    def __repr__(self):
        edition = self.edition
        try:
            assert edition in (1, 2)
            key_value_strings = []
            for argname in self.argnames:
                value = getattr(self, argname, None)
                assert isinstance(value, int)
                key_value_strings.append(f"{argname}={value}")
            inner_text = ", ".join(key_value_strings)
            result = f"GRIBCode({inner_text})"
        except Exception:
            # Invalid content somewhere : fall back on a default repr
            result = self._broken_repr()

        return result


class GRIBCode1(GenericConcreteGRIBCode):
    edition: int = 1
    table_version: int | None = None
    centre_number: int | None = None
    number: int | None = None
    argnames = ["edition", "table_version", "centre_number", "number"]
    _edition = 1  # Constructor argument should always match this


class GRIBCode2(GenericConcreteGRIBCode):
    edition: int = 2
    discipline: int | None = None
    category: int | None = None
    number: int | None = None
    argnames = ["edition", "discipline", "category", "number"]
    _edition = 2  # Constructor argument should always match this



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the BSD license.
# See LICENSE in the root of the repository for full licensing details.
"""
Unit tests for :meth:`iris_grib._save_rules.grid_definition_template_30`.

"""

# Import iris_grib.tests first so that some things can be initialised before
# importing anything else.
import iris_grib.tests as tests

import numpy as np

import iris.coords
from iris.coord_systems import GeogCS, LambertConformal
from iris.exceptions import TranslationError

from iris_grib._save_rules import grid_definition_template_30
from iris_grib.tests.unit.save_rules import GdtTestMixin


class FakeGribError(Exception):
    pass


class Test(tests.IrisGribTest, GdtTestMixin):
    def setUp(self):
        self.default_ellipsoid = GeogCS(
            semi_major_axis=6377563.396, semi_minor_axis=6356256.909
        )
        self.test_cube = self._make_test_cube()

        GdtTestMixin.setUp(self)

    def _make_test_cube(self, cs=None, x_points=None, y_points=None):
        # Create a cube with given properties, or minimal defaults.
        if cs is None:
            cs = self._default_coord_system()
        if x_points is None:
            x_points = self._default_x_points()
        if y_points is None:
            y_points = self._default_y_points()

        x_coord = iris.coords.DimCoord(
            x_points, "projection_x_coordinate", units="m", coord_system=cs
        )
        y_coord = iris.coords.DimCoord(
            y_points, "projection_y_coordinate", units="m", coord_system=cs
        )
        test_cube = iris.cube.Cube(np.zeros((len(y_points), len(x_points))))
        test_cube.add_dim_coord(y_coord, 0)
        test_cube.add_dim_coord(x_coord, 1)
        return test_cube

    def _default_coord_system(self, false_easting=0.0, false_northing=0.0):
        return LambertConformal(
            central_lat=39.0,
            central_lon=-96.0,
            false_easting=false_easting,
            false_northing=false_northing,
            secant_latitudes=(33, 45),
            ellipsoid=self.default_ellipsoid,
        )

    def test__template_number(self):
        grid_definition_template_30(self.test_cube, self.mock_grib)
        self._check_key("gridDefinitionTemplateNumber", 30)

    def test__shape_of_earth(self):
        grid_definition_template_30(self.test_cube, self.mock_grib)
        self._check_key("shapeOfTheEarth", 7)
        self._check_key("scaleFactorOfEarthMajorAxis", 0)
        self._check_key("scaledValueOfEarthMajorAxis", 6377563.396)
        self._check_key("scaleFactorOfEarthMinorAxis", 0)
        self._check_key("scaledValueOfEarthMinorAxis", 6356256.909)

    def test__grid_shape(self):
        test_cube = self._make_test_cube(x_points=np.arange(13), y_points=np.arange(6))
        grid_definition_template_30(test_cube, self.mock_grib)
        self._check_key("Nx", 13)
        self._check_key("Ny", 6)

    def test__grid_points(self):
        test_cube = self._make_test_cube(
            x_points=[1e6, 3e6, 5e6, 7e6], y_points=[4e6, 9e6]
        )
        grid_definition_template_30(test_cube, self.mock_grib)
        self._check_key("latitudeOfFirstGridPoint", 71676530)
        self._check_key("longitudeOfFirstGridPoint", 287218188)
        self._check_key("Dx", 2e9)
        self._check_key("Dy", 5e9)

    def test__template_specifics(self):
        grid_definition_template_30(self.test_cube, self.mock_grib)
        self._check_key("LaD", 39e6)
        self._check_key("LoV", 264e6)
        self._check_key("Latin1", 33e6)
        self._check_key("Latin2", 45e6)
        self._check_key("latitudeOfSouthernPole", 0)
        self._check_key("longitudeOfSouthernPole", 0)

    def test__scanmode(self):
        grid_definition_template_30(self.test_cube, self.mock_grib)
        self._check_scanmode(+1, +1)

    def test__scanmode_reverse(self):
        test_cube = self._make_test_cube(x_points=np.arange(7e6, 0, -1e6))
        grid_definition_template_30(test_cube, self.mock_grib)
        self._check_scanmode(-1, +1)

    def test_projection_centre(self):
        grid_definition_template_30(self.test_cube, self.mock_grib)
        self._check_key("projectionCentreFlag", 0)

    def test_projection_centre_south_pole(self):
        cs = LambertConformal(
            central_lat=39.0,
            central_lon=-96.0,
            false_easting=0.0,
            false_northing=0.0,
            secant_latitudes=(-33, -45),
            ellipsoid=self.default_ellipsoid,
        )
        test_cube = self._make_test_cube(cs=cs)
        grid_definition_template_30(test_cube, self.mock_grib)
        self._check_key("projectionCentreFlag", 128)

    def __fail_false_easting_northing(self, false_easting, false_northing):
        cs = self._default_coord_system(
            false_easting=false_easting, false_northing=false_northing
        )
        test_cube = self._make_test_cube(cs=cs)
        message = "Non-zero unsupported"
        with self.assertRaisesRegex(TranslationError, message):
            grid_definition_template_30(test_cube, self.mock_grib)

    def test__fail_false_easting(self):
        self.__fail_false_easting_northing(10.0, 0.0)

    def test__fail_false_northing(self):
        self.__fail_false_easting_northing(0.0, 10.0)

    def test__fail_false_easting_northing(self):
        self.__fail_false_easting_northing(10.0, 10.0)

    def __fail_irregular_coords(self, x=False, y=False):
        def irregular_coord(coord):
            coord = iris.coords.AuxCoord.from_coord(coord)
            coord.points[1] = coord.points[0]
            return coord

        test_cube = self._make_test_cube(
            # Make the Y dimension longer to make irregularity is possible.
            y_points=[7.0, 8.0, 9.0],
        )
        coord_x = test_cube.coord("projection_x_coordinate")
        coord_y = test_cube.coord("projection_y_coordinate")
        if x:
            test_cube.replace_coord(irregular_coord(coord_x))
        if y:
            test_cube.replace_coord(irregular_coord(coord_y))

        message = "Irregular coordinates not supported"
        with self.assertRaisesRegex(TranslationError, message):
            grid_definition_template_30(test_cube, self.mock_grib)

    def test__fail_irregular_x_coords(self):
        self.__fail_irregular_coords(x=True)

    def test__fail_irregular_y_coords(self):
        self.__fail_irregular_coords(y=True)

    def test__fail_irregular_coords(self):
        self.__fail_irregular_coords(x=True, y=True)


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the BSD license.
# See LICENSE in the root of the repository for full licensing details.
"""
Test function :func:`iris_grib._load_convert.vertical_coords`.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

from copy import deepcopy
from unittest import mock

from iris.coords import DimCoord
from iris.exceptions import TranslationError

from iris_grib._load_convert import vertical_coords
from iris_grib._load_convert import (
    _TYPE_OF_FIXED_SURFACE_MISSING as MISSING_SURFACE,
    _MDI as MISSING_LEVEL,
)


class Test(tests.IrisGribTest):
    def setUp(self):
        self.metadata = {
            "factories": [],
            "references": [],
            "standard_name": None,
            "long_name": None,
            "units": None,
            "attributes": {},
            "cell_methods": [],
            "dim_coords_and_dims": [],
            "aux_coords_and_dims": [],
        }

    def test_hybrid_factories(self):
        def func(section, metadata):
            return metadata["factories"].append(factory)

        metadata = deepcopy(self.metadata)
        section = {"NV": 1}
        this = "iris_grib._load_convert.hybrid_factories"
        factory = mock.sentinel.factory
        with mock.patch(this, side_effect=func) as hybrid_factories:
            vertical_coords(section, metadata)
            self.assertTrue(hybrid_factories.called)
            self.assertEqual(metadata["factories"], [factory])

    def test_no_first_fixed_surface(self):
        metadata = deepcopy(self.metadata)
        section = {
            "NV": 0,
            "typeOfFirstFixedSurface": MISSING_SURFACE,
            "scaledValueOfFirstFixedSurface": MISSING_LEVEL,
        }
        vertical_coords(section, metadata)
        self.assertEqual(metadata, self.metadata)

    def test_fixed_surface_type_1(self):
        metadata = deepcopy(self.metadata)
        section = {
            "NV": 0,
            "typeOfFirstFixedSurface": 1,
            "scaledValueOfFirstFixedSurface": 0,
            "scaleFactorOfFirstFixedSurface": 0,
            "typeOfSecondFixedSurface": 255,
        }
        vertical_coords(section, metadata)
        # No metadata change, as surfaceType=1 translates to "no vertical
        # coord" without error or warning.
        self.assertEqual(metadata, self.metadata)

    def test_unknown_first_fixed_surface_with_missing_scaled_value(self):
        this = "iris_grib._load_convert.options"
        with mock.patch("warnings.warn") as warn:
            with mock.patch(this) as options:
                for request_warning in [False, True]:
                    options.warn_on_unsupported = request_warning
                    metadata = deepcopy(self.metadata)
                    section = {
                        "NV": 0,
                        "typeOfFirstFixedSurface": 0,
                        "scaledValueOfFirstFixedSurface": MISSING_LEVEL,
                    }
                    # The call being tested.
                    vertical_coords(section, metadata)
                    self.assertEqual(metadata, self.metadata)
                    if request_warning:
                        self.assertEqual(len(warn.mock_calls), 1)
                        args, _ = warn.call_args
                        self.assertIn("surface with missing scaled value", args[0])
                    else:
                        self.assertEqual(len(warn.mock_calls), 0)

    def test_unknown_first_fixed_surface(self):
        metadata = deepcopy(self.metadata)
        expected = deepcopy(self.metadata)
        coord = DimCoord(600.0, attributes={"GRIB_fixed_surface_type": 106})
        expected["aux_coords_and_dims"].append((coord, None))

        section = {
            "NV": 0,
            "typeOfFirstFixedSurface": 106,
            "scaledValueOfFirstFixedSurface": 600,
            "scaleFactorOfFirstFixedSurface": 0,
            "typeOfSecondFixedSurface": MISSING_SURFACE,
        }
        vertical_coords(section, metadata)
        self.assertEqual(metadata, expected)

    def test_unknown_first_fixed_surface_with_second_fixed_surface(self):
        metadata = deepcopy(self.metadata)
        expected = deepcopy(self.metadata)
        coord = DimCoord(
            9000.0, bounds=[18000, 0], attributes={"GRIB_fixed_surface_type": 108}
        )
        expected["aux_coords_and_dims"].append((coord, None))

        section = {
            "NV": 0,
            "typeOfFirstFixedSurface": 108,
            "scaledValueOfFirstFixedSurface": 18000,
            "scaleFactorOfFirstFixedSurface": 0,
            "typeOfSecondFixedSurface": 108,
            "scaledValueOfSecondFixedSurface": 0,
            "scaleFactorOfSecondFixedSurface": 0,
        }
        vertical_coords(section, metadata)
        self.assertEqual(metadata, expected)

    def test_pressure_with_no_second_fixed_surface(self):
        metadata = deepcopy(self.metadata)
        section = {
            "NV": 0,
            "typeOfFirstFixedSurface": 100,  # pressure / Pa
            "scaledValueOfFirstFixedSurface": 10,
            "scaleFactorOfFirstFixedSurface": 1,
            "typeOfSecondFixedSurface": MISSING_SURFACE,
        }
        vertical_coords(section, metadata)
        coord = DimCoord(1.0, long_name="pressure", units="Pa")
        expected = deepcopy(self.metadata)
        expected["aux_coords_and_dims"].append((coord, None))
        self.assertEqual(metadata, expected)

    def test_height_with_no_second_fixed_surface(self):
        metadata = deepcopy(self.metadata)
        section = {
            "NV": 0,
            "typeOfFirstFixedSurface": 103,  # height / m
            "scaledValueOfFirstFixedSurface": 100,
            "scaleFactorOfFirstFixedSurface": 2,
            "typeOfSecondFixedSurface": MISSING_SURFACE,
        }
        vertical_coords(section, metadata)
        coord = DimCoord(1.0, long_name="height", units="m")
        expected = deepcopy(self.metadata)
        expected["aux_coords_and_dims"].append((coord, None))
        self.assertEqual(metadata, expected)

    def test_different_fixed_surfaces(self):
        section = {
            "NV": 0,
            "typeOfFirstFixedSurface": 100,
            "scaledValueOfFirstFixedSurface": 10,
            "scaleFactorOfFirstFixedSurface": 1,
            "typeOfSecondFixedSurface": 0,
        }
        emsg = "different types of first and second fixed surface"
        with self.assertRaisesRegex(TranslationError, emsg):
            vertical_coords(section, None)

    def test_same_fixed_surfaces_missing_second_scaled_value(self):
        section = {
            "NV": 0,
            "typeOfFirstFixedSurface": 100,
            "scaledValueOfFirstFixedSurface": 10,
            "scaleFactorOfFirstFixedSurface": 1,
            "typeOfSecondFixedSurface": 100,
            "scaledValueOfSecondFixedSurface": MISSING_LEVEL,
        }
        emsg = "missing scaled value of second fixed surface"
        with self.assertRaisesRegex(TranslationError, emsg):
            vertical_coords(section, None)

    def test_pressure_with_second_fixed_surface(self):
        metadata = deepcopy(self.metadata)
        section = {
            "NV": 0,
            "typeOfFirstFixedSurface": 100,
            "scaledValueOfFirstFixedSurface": 10,
            "scaleFactorOfFirstFixedSurface": 1,
            "typeOfSecondFixedSurface": 100,
            "scaledValueOfSecondFixedSurface": 30,
            "scaleFactorOfSecondFixedSurface": 1,
        }
        vertical_coords(section, metadata)
        coord = DimCoord(2.0, long_name="pressure", units="Pa", bounds=[1.0, 3.0])
        expected = deepcopy(self.metadata)
        expected["aux_coords_and_dims"].append((coord, None))
        self.assertEqual(metadata, expected)

    def test_height_with_second_fixed_surface(self):
        metadata = deepcopy(self.metadata)
        section = {
            "NV": 0,
            "typeOfFirstFixedSurface": 103,
            "scaledValueOfFirstFixedSurface": 1000,
            "scaleFactorOfFirstFixedSurface": 2,
            "typeOfSecondFixedSurface": 103,
            "scaledValueOfSecondFixedSurface": 3000,
            "scaleFactorOfSecondFixedSurface": 2,
        }
        vertical_coords(section, metadata)
        coord = DimCoord(20.0, long_name="height", units="m", bounds=[10.0, 30.0])
        expected = deepcopy(self.metadata)
        expected["aux_coords_and_dims"].append((coord, None))
        self.assertEqual(metadata, expected)


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the BSD license.
# See LICENSE in the root of the repository for full licensing details.
"""
Unit tests for :meth:`iris_grib._save_rules.grid_definition_template_20`.

"""

# Import iris_grib.tests first so that some things can be initialised before
# importing anything else.
import iris_grib.tests as tests

from iris.coord_systems import GeogCS, PolarStereographic, Stereographic
from iris.coords import AuxCoord
from iris.exceptions import TranslationError
import numpy as np

from iris_grib._save_rules import grid_definition_template_20
from iris_grib.tests.unit.save_rules import GdtTestMixin


class Test(tests.IrisGribTest, GdtTestMixin):
    def setUp(self):
        self.default_ellipsoid = GeogCS(semi_major_axis=6371200.0)
        self.stereo_test_cube = self._make_test_cube(coord_units="m")

        GdtTestMixin.setUp(self)

    def _default_coord_system(self, false_easting=0, false_northing=0):
        return PolarStereographic(
            90.0,
            0,
            false_easting=false_easting,
            false_northing=false_northing,
            ellipsoid=self.default_ellipsoid,
        )

    def test__template_number(self):
        grid_definition_template_20(self.stereo_test_cube, self.mock_grib)
        self._check_key("gridDefinitionTemplateNumber", 20)

    def test__shape_of_earth(self):
        grid_definition_template_20(self.stereo_test_cube, self.mock_grib)
        self._check_key("shapeOfTheEarth", 1)
        self._check_key("scaleFactorOfRadiusOfSphericalEarth", 0)
        self._check_key("scaleFactorOfEarthMajorAxis", 0)
        self._check_key("scaledValueOfEarthMajorAxis", 0)
        self._check_key("scaleFactorOfEarthMinorAxis", 0)
        self._check_key("scaledValueOfEarthMinorAxis", 0)

    def test__grid_shape(self):
        stereo_test_cube = self._make_test_cube(
            x_points=np.arange(13), y_points=np.arange(6), coord_units="m"
        )
        grid_definition_template_20(stereo_test_cube, self.mock_grib)
        self._check_key("Nx", 13)
        self._check_key("Ny", 6)

    def test__grid_points(self):
        stereo_test_cube = self._make_test_cube(
            x_points=[1e6, 3e6, 5e6, 7e6], y_points=[4e6, 9e6], coord_units="m"
        )
        grid_definition_template_20(stereo_test_cube, self.mock_grib)
        self._check_key("latitudeOfFirstGridPoint", 54139565)
        self._check_key("longitudeOfFirstGridPoint", 165963757)
        self._check_key("Dx", 2e9)
        self._check_key("Dy", 5e9)

    def test__template_specifics(self):
        grid_definition_template_20(self.stereo_test_cube, self.mock_grib)
        self._check_key("LaD", 90e6)
        self._check_key("LoV", 0)

    def test__scanmode(self):
        grid_definition_template_20(self.stereo_test_cube, self.mock_grib)
        self._check_scanmode(+1, +1)

    def test__scanmode_reverse(self):
        stereo_test_cube = self._make_test_cube(
            x_points=np.arange(7e6, 0, -1e6), coord_units="m"
        )
        grid_definition_template_20(stereo_test_cube, self.mock_grib)
        self._check_scanmode(-1, +1)

    def test_projection_centre(self):
        grid_definition_template_20(self.stereo_test_cube, self.mock_grib)
        self._check_key("projectionCentreFlag", 0)

    def test_projection_centre_south_pole(self):
        cs = PolarStereographic(-90.0, 0, ellipsoid=self.default_ellipsoid)
        stereo_test_cube = self._make_test_cube(cs=cs, coord_units="m")
        grid_definition_template_20(stereo_test_cube, self.mock_grib)
        self._check_key("projectionCentreFlag", 128)

    def test_projection_centre_south_pole_parent(self):
        # Saving should be able to handle either class (PolarStereographic
        #  being a subclass of Stereographic).
        cs = Stereographic(-90.0, 0, ellipsoid=self.default_ellipsoid)
        stereo_test_cube = self._make_test_cube(cs=cs, coord_units="m")
        grid_definition_template_20(stereo_test_cube, self.mock_grib)
        self._check_key("projectionCentreFlag", 128)

    def test_projection_centre_bad(self):
        cs = Stereographic(0, 0, ellipsoid=self.default_ellipsoid)
        stereo_test_cube = self._make_test_cube(cs=cs, coord_units="m")
        exp_emsg = "must be 90.0 or -90.0"
        with self.assertRaisesRegex(TranslationError, exp_emsg):
            grid_definition_template_20(stereo_test_cube, self.mock_grib)

    def __fail_false_easting_northing(self, false_easting, false_northing):
        cs = self._default_coord_system(
            false_easting=false_easting, false_northing=false_northing
        )
        test_cube = self._make_test_cube(cs=cs)
        message = "Non-zero unsupported"
        with self.assertRaisesRegex(TranslationError, message):
            grid_definition_template_20(test_cube, self.mock_grib)

    def test__fail_false_easting(self):
        self.__fail_false_easting_northing(10.0, 0.0)

    def test__fail_false_northing(self):
        self.__fail_false_easting_northing(0.0, 10.0)

    def test__fail_false_easting_northing(self):
        self.__fail_false_easting_northing(10.0, 10.0)

    def __fail_irregular_coords(self, x=False, y=False):
        def irregular_coord(coord):
            coord = AuxCoord.from_coord(coord)
            coord.points[1] = coord.points[0]
            return coord

        test_cube = self._make_test_cube(
            # Make the Y dimension longer to make irregularity is possible.
            y_points=[7.0, 8.0, 9.0],
            coord_units="m",
        )
        coord_lon = test_cube.coord("longitude")
        coord_lat = test_cube.coord("latitude")
        if x:
            test_cube.replace_coord(irregular_coord(coord_lon))
        if y:
            test_cube.replace_coord(irregular_coord(coord_lat))

        message = "Irregular coordinates not supported"
        with self.assertRaisesRegex(TranslationError, message):
            grid_definition_template_20(test_cube, self.mock_grib)

    def test__fail_irregular_x_coords(self):
        self.__fail_irregular_coords(x=True)

    def test__fail_irregular_y_coords(self):
        self.__fail_irregular_coords(y=True)

    def test__fail_irregular_coords(self):
        self.__fail_irregular_coords(x=True, y=True)

    def test__fail_non_identical_lats(self):
        coord_system = PolarStereographic(
            90.0,
            0,
            true_scale_lat=60.0,
            ellipsoid=self.default_ellipsoid,
        )
        test_cube = self._make_test_cube(cs=coord_system, coord_units="m")
        message = "only write a GRIB Template 3.20 file where these are identical"
        with self.assertRaisesRegex(TranslationError, message):
            grid_definition_template_20(test_cube, self.mock_grib)

    def test__fail_scale_factor(self):
        coord_system = PolarStereographic(
            90.0,
            0,
            scale_factor_at_projection_origin=0.5,
            ellipsoid=self.default_ellipsoid,
        )
        test_cube = self._make_test_cube(cs=coord_system, coord_units="m")
        message = "cannot write scale_factor_at_projection_origin"
        with self.assertRaisesRegex(TranslationError, message):
            grid_definition_template_20(test_cube, self.mock_grib)


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the BSD license.
# See LICENSE in the root of the repository for full licensing details.
"""Test function :func:`iris_grib._load_convert.grib2_convert`."""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

import copy
from unittest import mock

import iris_grib
from iris_grib._load_convert import grib2_convert
from iris_grib.tests.unit import _make_test_message


class Test(tests.IrisGribTest):
    def setUp(self):
        this = "iris_grib._load_convert"
        self.patch("{}.reference_time_coord".format(this), return_value=None)
        self.patch("{}.grid_definition_section".format(this))
        self.patch("{}.product_definition_section".format(this))
        self.patch("{}.data_representation_section".format(this))
        self.patch("{}.bitmap_section".format(this))

    def test(self):
        sections = [
            {"discipline": mock.sentinel.discipline},  # section 0
            {
                "centre": "ecmf",  # section 1
                "tablesVersion": mock.sentinel.tablesVersion,
            },
            None,  # section 2
            mock.sentinel.grid_definition_section,  # section 3
            mock.sentinel.product_definition_section,  # section 4
            mock.sentinel.data_representation_section,  # section 5
            mock.sentinel.bitmap_section,
        ]  # section 6
        field = _make_test_message(sections)
        metadata = {
            "factories": [],
            "references": [],
            "standard_name": None,
            "long_name": None,
            "units": None,
            "attributes": {},
            "cell_methods": [],
            "dim_coords_and_dims": [],
            "aux_coords_and_dims": [],
        }
        expected = copy.deepcopy(metadata)
        centre = "European Centre for Medium Range Weather Forecasts"
        expected["attributes"] = {"centre": centre}
        # The call being tested.
        grib2_convert(field, metadata)
        self.assertEqual(metadata, expected)
        this = iris_grib._load_convert
        this.reference_time_coord.assert_called_with(sections[1])
        this.grid_definition_section.assert_called_with(sections[3], expected)
        args = (
            sections[4],
            expected,
            sections[0]["discipline"],
            sections[1]["tablesVersion"],
            None,
        )
        this.product_definition_section.assert_called_with(*args)


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the BSD license.
# See LICENSE in the root of the repository for full licensing details.
"""
Unit tests for `iris_grib.message.Section`.

"""

# Import iris_grib.tests first so that some things can be initialised before
# importing anything else.
import iris_grib.tests as tests

import eccodes
import numpy as np

from iris_grib.message import Section


@tests.skip_data
class Test___getitem__(tests.IrisGribTest):
    def setUp(self):
        filename = tests.get_data_path(("GRIB", "uk_t", "uk_t.grib2"))
        with open(filename, "rb") as grib_fh:
            self.grib_id = eccodes.codes_new_from_file(
                grib_fh, eccodes.CODES_PRODUCT_GRIB
            )

    def test_scalar(self):
        section = Section(self.grib_id, None, ["Ni"])
        self.assertEqual(section["Ni"], 47)

    def test_array(self):
        section = Section(self.grib_id, None, ["codedValues"])
        codedValues = section["codedValues"]
        self.assertEqual(codedValues.shape, (1551,))
        self.assertArrayAlmostEqual(
            codedValues[:3], [-1.78140259, -1.53140259, -1.28140259]
        )

    def test_typeOfFirstFixedSurface(self):
        section = Section(self.grib_id, None, ["typeOfFirstFixedSurface"])
        self.assertEqual(section["typeOfFirstFixedSurface"], 100)

    def test_numberOfSection(self):
        n = 4
        section = Section(self.grib_id, n, ["numberOfSection"])
        self.assertEqual(section["numberOfSection"], n)

    def test_invalid(self):
        section = Section(self.grib_id, None, ["Ni"])
        with self.assertRaisesRegex(KeyError, "Nii"):
            section["Nii"]


@tests.skip_data
class Test__getitem___pdt_31(tests.IrisGribTest):
    def setUp(self):
        filename = tests.get_data_path(("GRIB", "umukv", "ukv_chan9.grib2"))
        with open(filename, "rb") as grib_fh:
            self.grib_id = eccodes.codes_new_from_file(
                grib_fh, eccodes.CODES_PRODUCT_GRIB
            )
        self.keys = [
            "satelliteSeries",
            "satelliteNumber",
            "instrumentType",
            "scaleFactorOfCentralWaveNumber",
            "scaledValueOfCentralWaveNumber",
        ]

    def test_array(self):
        section = Section(self.grib_id, None, self.keys)
        for key in self.keys:
            value = section[key]
            self.assertIsInstance(value, np.ndarray)
            self.assertEqual(value.shape, (1,))


@tests.skip_data
class Test_get_computed_key(tests.IrisGribTest):
    def test_gdt40_computed(self):
        fname = tests.get_data_path(("GRIB", "gaussian", "regular_gg.grib2"))
        with open(fname, "rb") as grib_fh:
            self.grib_id = eccodes.codes_new_from_file(
                grib_fh, eccodes.CODES_PRODUCT_GRIB
            )
            section = Section(self.grib_id, None, [])
        latitudes = section.get_computed_key("latitudes")
        self.assertTrue(88.55 < latitudes[0] < 88.59)


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the BSD license.
# See LICENSE in the root of the repository for full licensing details.
"""
Tests for `iris_grib._load_convert.product_definition_template_31`.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

from unittest import mock

from iris_grib.tests.unit.load_convert import empty_metadata

from iris_grib._load_convert import product_definition_template_31


def section_4():
    # Also needed for test_product_definition_section.py.
    series = mock.sentinel.satelliteSeries
    number = mock.sentinel.satelliteNumber
    instrument = mock.sentinel.instrumentType
    return {
        "NB": 1,
        "satelliteSeries": series,
        "satelliteNumber": number,
        "instrumentType": instrument,
        "scaleFactorOfCentralWaveNumber": 1,
        "scaledValueOfCentralWaveNumber": 12,
    }


class Test(tests.IrisGribTest):
    def setUp(self):
        self.patch("warnings.warn")
        self.satellite_common_patch = self.patch(
            "iris_grib._load_convert.satellite_common"
        )
        self.generating_process_patch = self.patch(
            "iris_grib._load_convert.generating_process"
        )

    def test(self):
        # Prepare the arguments.
        rt_coord = mock.sentinel.observation_time
        section = section_4()

        # Call the function.
        metadata = empty_metadata()
        product_definition_template_31(section, metadata, rt_coord)

        # Check that 'satellite_common' was called.
        self.assertEqual(self.satellite_common_patch.call_count, 1)
        # Check that 'generating_process' was called.
        self.assertEqual(self.generating_process_patch.call_count, 1)
        # Check that the scalar time coord was added in.
        self.assertIn((rt_coord, None), metadata["aux_coords_and_dims"])


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the BSD license.
# See LICENSE in the root of the repository for full licensing details.
"""
Test function :func:`iris_grib._load_convert.time_range_unit.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

from cf_units import Unit
from iris.exceptions import TranslationError

from iris_grib._load_convert import time_range_unit


class Test(tests.IrisGribTest):
    def setUp(self):
        self.unit_by_indicator = {
            0: Unit("minutes"),
            1: Unit("hours"),
            2: Unit("days"),
            10: Unit("3 hours"),
            11: Unit("6 hours"),
            12: Unit("12 hours"),
            13: Unit("seconds"),
        }

    def test_units(self):
        for indicator, unit in self.unit_by_indicator.items():
            result = time_range_unit(indicator)
            self.assertEqual(result, unit)

    def test_bad_indicator(self):
        emsg = "unsupported time range"
        with self.assertRaisesRegex(TranslationError, emsg):
            time_range_unit(-1)


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the BSD license.
# See LICENSE in the root of the repository for full licensing details.
"""
Unit tests for the mod:`iris_grib.grib_phenom_translation` module.

Carried over from old iris/tests/test_grib_phenom_translation.py.
Code is out of step with current test conventions and standards.

"""

# Import iris_grib.tests first so that some things can be initialised before
# importing anything else.
import iris_grib.tests as tests

import cf_units

import iris_grib.grib_phenom_translation as gptx
from iris_grib.grib_phenom_translation import GRIBCode


class TestGribLookupTableType(tests.IrisTest):
    def test_lookuptable_type(self):
        ll = gptx._LookupTable([("a", 1), ("b", 2)])
        self.assertEqual(1, ll["a"])
        self.assertIsNone(ll["q"])
        ll["q"] = 15
        self.assertEqual(15, ll["q"])
        ll["q"] = 15
        self.assertEqual(15, ll["q"])
        with self.assertRaises(KeyError):
            ll["q"] = 7
        del ll["q"]
        ll["q"] = 7
        self.assertEqual(7, ll["q"])


class TestGribPhenomenonLookup(tests.IrisTest):
    def test_grib1_cf_lookup(self):
        def check_grib1_cf(
            param,
            standard_name,
            long_name,
            units,
            height=None,
            t2version=128,
            centre=98,
            expect_none=False,
        ):
            a_cf_unit = cf_units.Unit(units)
            cfdata = gptx.grib1_phenom_to_cf_info(
                param_number=param, table2_version=t2version, centre_number=centre
            )
            if expect_none:
                self.assertIsNone(cfdata)
            else:
                self.assertEqual(cfdata.standard_name, standard_name)
                self.assertEqual(cfdata.long_name, long_name)
                self.assertEqual(cfdata.units, a_cf_unit)
                if height is None:
                    self.assertIsNone(cfdata.set_height)
                else:
                    self.assertEqual(cfdata.set_height, float(height))

        check_grib1_cf(165, "x_wind", None, "m s-1", 10.0)
        check_grib1_cf(168, "dew_point_temperature", None, "K", 2)
        check_grib1_cf(130, "air_temperature", None, "K")
        check_grib1_cf(235, None, "grib_skin_temperature", "K")
        check_grib1_cf(
            235, None, "grib_skin_temperature", "K", t2version=9999, expect_none=True
        )
        check_grib1_cf(
            235, None, "grib_skin_temperature", "K", centre=9999, expect_none=True
        )
        check_grib1_cf(9999, None, "grib_skin_temperature", "K", expect_none=True)

    def test_grib2_cf_lookup(self):
        def check_grib2_cf(
            discipline,
            category,
            number,
            standard_name,
            long_name,
            units,
            expect_none=False,
        ):
            a_cf_unit = cf_units.Unit(units)
            cfdata = gptx.grib2_phenom_to_cf_info(
                param_discipline=discipline,
                param_category=category,
                param_number=number,
            )
            if expect_none:
                self.assertIsNone(cfdata)
            else:
                self.assertEqual(cfdata.standard_name, standard_name)
                self.assertEqual(cfdata.long_name, long_name)
                self.assertEqual(cfdata.units, a_cf_unit)

        # These should work
        check_grib2_cf(0, 0, 2, "air_potential_temperature", None, "K")
        check_grib2_cf(0, 19, 1, None, "grib_physical_atmosphere_albedo", "%")
        check_grib2_cf(2, 0, 2, "soil_temperature", None, "K")
        check_grib2_cf(10, 2, 0, "sea_ice_area_fraction", None, 1)
        check_grib2_cf(2, 0, 0, "land_area_fraction", None, 1)
        check_grib2_cf(0, 19, 1, None, "grib_physical_atmosphere_albedo", "%")
        check_grib2_cf(
            0, 1, 64, "atmosphere_mass_content_of_water_vapor", None, "kg m-2"
        )
        check_grib2_cf(2, 0, 7, "surface_altitude", None, "m")

        # These should fail
        check_grib2_cf(9999, 2, 0, "sea_ice_area_fraction", None, 1, expect_none=True)
        check_grib2_cf(10, 9999, 0, "sea_ice_area_fraction", None, 1, expect_none=True)
        check_grib2_cf(10, 2, 9999, "sea_ice_area_fraction", None, 1, expect_none=True)

    def test_cf_grib2_lookup(self):
        def check_cf_grib2(
            standard_name,
            long_name,
            discipline,
            category,
            number,
            units,
            expect_none=False,
        ):
            a_cf_unit = cf_units.Unit(units)
            gribdata = gptx.cf_phenom_to_grib2_info(standard_name, long_name)
            if expect_none:
                self.assertIsNone(gribdata)
            else:
                self.assertEqual(gribdata.discipline, discipline)
                self.assertEqual(gribdata.category, category)
                self.assertEqual(gribdata.number, number)
                self.assertEqual(gribdata.units, a_cf_unit)

        # These should work
        check_cf_grib2("sea_surface_temperature", None, 10, 3, 0, "K")
        check_cf_grib2("air_temperature", None, 0, 0, 0, "K")
        check_cf_grib2("soil_temperature", None, 2, 0, 2, "K")
        check_cf_grib2("land_area_fraction", None, 2, 0, 0, "1")
        check_cf_grib2("land_binary_mask", None, 2, 0, 0, "1")
        check_cf_grib2(
            "atmosphere_mass_content_of_water_vapor", None, 0, 1, 64, "kg m-2"
        )
        check_cf_grib2("surface_altitude", None, 2, 0, 7, "m")

        # These should fail
        check_cf_grib2("air_temperature", "user_long_UNRECOGNISED", 0, 0, 0, "K")
        check_cf_grib2(
            "air_temperature_UNRECOGNISED", None, 0, 0, 0, "K", expect_none=True
        )
        check_cf_grib2(None, "user_long_UNRECOGNISED", 0, 0, 0, "K", expect_none=True)
        check_cf_grib2(None, "precipitable_water", 0, 1, 3, "kg m-2")
        check_cf_grib2(
            "invalid_unknown", "precipitable_water", 0, 1, 3, "kg m-2", expect_none=True
        )
        check_cf_grib2(None, None, 0, 0, 0, "", expect_none=True)


class TestGRIBcode(tests.IrisTest):
    # GRIBCode is basically a namedtuple, so not all behaviour needs testing.
    # However, creation is a bit special so exercise all those cases.

    # TODO: convert to pytest + replace duplications with parameterisation
    #   (mostly grib1/grib2, but also in one case str/repr)
    def test_create_from_keys__grib2(self):
        gribcode = GRIBCode(edition=2, discipline=7, category=4, number=199)
        self.assertEqual(gribcode.edition, 2)
        self.assertEqual(gribcode.discipline, 7)
        self.assertEqual(gribcode.category, 4)
        self.assertEqual(gribcode.number, 199)

    def test_create_from_keys__grib1(self):
        gribcode = GRIBCode(edition=1, table_version=7, centre_number=4, number=199)
        self.assertEqual(gribcode.edition, 1)
        self.assertEqual(gribcode.table_version, 7)
        self.assertEqual(gribcode.centre_number, 4)
        self.assertEqual(gribcode.number, 199)

    def test_create_from_args__grib2(self):
        gribcode = GRIBCode(2, 3, 12, 99)
        self.assertEqual(gribcode.edition, 2)
        self.assertEqual(gribcode.discipline, 3)
        self.assertEqual(gribcode.category, 12)
        self.assertEqual(gribcode.number, 99)

    def test_create_from_args__grib1(self):
        gribcode = GRIBCode(1, 3, 12, 99)
        self.assertEqual(gribcode.edition, 1)
        self.assertEqual(gribcode.table_version, 3)
        self.assertEqual(gribcode.centre_number, 12)
        self.assertEqual(gribcode.number, 99)

    def check_create_is_copy(self, edition):
        gribcode1 = GRIBCode(edition, 3, 12, 99)
        gribcode2 = GRIBCode(edition, 3, 12, 99)
        self.assertEqual(gribcode1, gribcode2)
        self.assertIsNot(gribcode1, gribcode2)

    def test_create_is_copy__grib1(self):
        self.check_create_is_copy(edition=1)

    def test_create_is_copy__grib2(self):
        self.check_create_is_copy(edition=2)

    def check_create_from_gribcode(self, edition):
        gribcode1 = GRIBCode((edition, 3, 2, 1))
        gribcode2 = GRIBCode(gribcode1)
        self.assertEqual(gribcode1, gribcode2)
        # NOTE: *not* passthrough : it creates a copy
        # (though maybe not too significant, as it is immutable anyway?)
        self.assertIsNot(gribcode1, gribcode2)

    def test_create_from_gribcode__grib1(self):
        self.check_create_from_gribcode(edition=1)

    def test_create_from_gribcode__grib2(self):
        self.check_create_from_gribcode(edition=2)

    def check_create_from_string(self, edition):
        gribcode = GRIBCode(f"xxx{edition}xs-34 -5,678qqqq")
        # NOTE: args 2 and 3 are *not* negative.
        self.assertEqual(gribcode, GRIBCode(edition, 34, 5, 678))

    def test_create_from_string__grib1(self):
        self.check_create_from_string(edition=1)

    def test_create_from_string__grib2(self):
        self.check_create_from_string(edition=2)

    def check_create_from_own_string(self, string_function, edition):
        # Check that GRIBCode string reprs are valid as create arguments.
        gribcode = GRIBCode(edition, 17, 94, 231)
        grib_param_string = string_function(gribcode)
        newcode = GRIBCode(grib_param_string)
        self.assertEqual(newcode, gribcode)

    def test_create_from_own_string__str__grib1(self):
        self.check_create_from_own_string(str, edition=1)

    def test_create_from_own_string__str__grib2(self):
        self.check_create_from_own_string(str, edition=2)

    def test_create_from_own_string__repr__grib1(self):
        self.check_create_from_own_string(repr, edition=1)

    def test_create_from_own_string__repr__grib2(self):
        self.check_create_from_own_string(repr, edition=2)

    def check_create_from_tuple(self, edition):
        gribcode = GRIBCode((edition, 3, 2, 1))
        expected = GRIBCode(edition, 3, 2, 1)
        self.assertEqual(expected, gribcode)

    def test_create_from_tuple__grib1(self):
        self.check_create_from_tuple(edition=1)

    def test_create_from_tuple__grib2(self):
        self.check_create_from_tuple(edition=2)

    def test_create_bad_nargs(self):
        # Between 1 and 4 args is not invalid call syntax, but it should fail.
        msg = (
            "Cannot create.* from 2 arguments.*"
            r"GRIBCode\(\(1, 2\)\).*"
            "expects either 1 or 4 arguments"
        )
        with self.assertRaisesRegex(ValueError, msg):
            GRIBCode(1, 2)

    def test_create_bad_single_arg_None(self):
        msg = (
            "Cannot create GRIBCode from 0 arguments.*"
            r"GRIBCode\(\(\)\).*"
            "expects either 1 or 4 arguments"
        )
        with self.assertRaisesRegex(ValueError, msg):
            GRIBCode(None)

    def test_create_bad_single_arg_empty_string(self):
        msg = (
            "Invalid argument for GRIBCode creation.*"
            r"GRIBCode\(''\).*"
            "requires 4 numbers, separated by non-numerals"
        )
        with self.assertRaisesRegex(ValueError, msg):
            GRIBCode("")

    def test_create_bad_single_arg_nonums(self):
        msg = (
            "Invalid argument for GRIBCode creation.*"
            r"GRIBCode\('saas- dsa- '\).*"
            "requires 4 numbers, separated by non-numerals"
        )
        with self.assertRaisesRegex(ValueError, msg):
            GRIBCode("saas- dsa- ")

    def test_create_bad_single_arg_less_than_4_nums(self):
        msg = (
            "Invalid argument for GRIBCode creation.*"
            r"GRIBCode\('1,2,3'\).*"
            "requires 4 numbers, separated by non-numerals"
        )
        with self.assertRaisesRegex(ValueError, msg):
            GRIBCode("1,2,3")

    def test_create_bad_single_arg_number(self):
        msg = (
            "Invalid argument for GRIBCode creation.*"
            r"GRIBCode\('4'\).*"
            "requires 4 numbers, separated by non-numerals"
        )
        with self.assertRaisesRegex(ValueError, msg):
            GRIBCode(4)

    def test_create_bad_single_arg_single_numeric(self):
        msg = (
            "Invalid argument for GRIBCode creation.*"
            r"GRIBCode\('44'\).*"
            "requires 4 numbers, separated by non-numerals"
        )
        with self.assertRaisesRegex(ValueError, msg):
            GRIBCode("44")

    def test_create_string_more_than_4_nums(self):
        # Note: does not error, just discards the extra.
        gribcode = GRIBCode("1,2,3,4,5,6,7,8")
        self.assertEqual(gribcode, GRIBCode(1, 2, 3, 4))

    def check__str__(self, edition):
        result = str(GRIBCode(edition, 17, 3, 123))
        arg1_char = {1: "t", 2: "d"}[edition]
        expected = f"GRIB{edition}:{arg1_char}017c003n123"
        self.assertEqual(expected, result)

    def test__str__grib1(self):
        self.check__str__(edition=1)

    def test__str__grib2(self):
        self.check__str__(edition=2)

    def check__repr__(self, edition):
        result = repr(GRIBCode(edition, 17, 3, 123))
        if edition == 1:
            expected = (
                "GRIBCode(edition=1, table_version=17, " "centre_number=3, number=123)"
            )
        elif edition == 2:
            expected = "GRIBCode(edition=2, discipline=17, " "category=3, number=123)"
        self.assertEqual(result, expected)

    def test__repr__grib1(self):
        self.check__repr__(edition=1)

    def test__repr__grib2(self):
        self.check__repr__(edition=2)

    def test_bad_content__str_repr__badedition(self):
        gribcode = GRIBCode(2, 11, 12, 13)
        gribcode.edition = 77
        str_result = str(gribcode)
        expected = (
            "<GRIBCode2 with invalid content: "
            "{'edition': 77, 'discipline': 11, 'category': 12, 'number': 13}>"
        )
        self.assertEqual(expected, str_result)
        repr_result = repr(gribcode)
        self.assertEqual(str_result, repr_result)

    def test_bad_content__str_repr__badmembervalue(self):
        gribcode = GRIBCode(2, 11, 12, 13)
        gribcode.discipline = None
        str_result = str(gribcode)
        expected = (
            "<GRIBCode2 with invalid content: "
            "{'edition': 2, 'discipline': None, 'category': 12, 'number': 13}>"
        )
        self.assertEqual(expected, str_result)
        repr_result = repr(gribcode)
        self.assertEqual(str_result, repr_result)

    def test_bad_content__str_repr__missingmember(self):
        gribcode = GRIBCode(2, 11, 12, 13)
        del gribcode.category
        str_result = str(gribcode)
        expected = (
            "<GRIBCode2 with invalid content: "
            "{'edition': 2, 'discipline': 11, 'number': 13}>"
        )
        self.assertEqual(expected, str_result)
        repr_result = repr(gribcode)
        self.assertEqual(str_result, repr_result)

    def test_bad_create__invalid_edition(self):
        with self.assertRaisesRegex(ValueError, "Invalid grib edition"):
            GRIBCode(77, 1, 2, 3)

    def test_bad_create__arg_and_kwarg(self):
        msg = "Keyword 'number'=7 is not compatible with a 4th argument."
        with self.assertRaisesRegex(ValueError, msg):
            GRIBCode(1, 2, 3, 4, number=7)


if __name__ == "__main__":
    tests.main()



# Copyright iris-grib contributors
#
# This file is part of iris-grib and is released under the BSD license.
# See LICENSE in the root of the repository for full licensing details.
"""
Test function :func:`iris_grib._load_convert.reference_time_coord.

Reference Code Table 1.2.

"""

# import iris_grib.tests first so that some things can be initialised
# before importing anything else.
import iris_grib.tests as tests

from copy import deepcopy
from datetime import datetime

from cf_units import CALENDAR_GREGORIAN, Unit

from iris.coords import DimCoord
from iris.exceptions import TranslationError

from iris_grib._load_convert import reference_time_coord


class Test(tests.IrisGribTest):
    def setUp(self):
        self.section = {
            "year": 2007,
            "month": 1,
            "day": 15,
            "hour": 0,
            "minute": 3,
            "second": 0,
        }
        self.unit = Unit("hours since epoch", calendar=CALENDAR_GREGORIAN)
        dt = datetime(
            self.section["year"],
            self.section["month"],
            self.section["day"],
            self.section["hour"],
            self.section["minute"],
            self.section["second"],
        )
        self.point = self.unit.date2num(dt)

    def _check(self, section, standard_name=None):
        expected = DimCoord(self.point, standard_name=standard_name, units=self.unit)
        # The call being tested.
        coord = reference_time_coord(section)
        self.assertEqual(coord, expected)

    def test_start_of_forecast__0(self):
        section = deepcopy(self.section)
        section["significanceOfReferenceTime"] = 0
        self._check(section, "forecast_reference_time")

    def test_start_of_forecast__1(self):
        section = deepcopy(self.section)
        section["significanceOfReferenceTime"] = 1
        self._check(section, "forecast_reference_time")

    def test_observation_time(self):
        section = deepcopy(self.section)
        section["significanceOfReferenceTime"] = 3
        self._check(section, "time")

    def test_unknown_significance(self):
        section = deepcopy(self.section)
        section["significanceOfReferenceTime"] = 5
        emsg = "unsupported significance"
        with self.assertRaisesRegex(TranslationError, emsg):
            self._check(section)


if __name__ == "__main__":
    tests.main()
